package com.ef_prime.rflow;

import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.net.URL;

import javax.help.CSH;
import javax.help.HelpBroker;
import javax.help.HelpSet;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JToolBar;
import javax.swing.KeyStroke;
import javax.swing.filechooser.FileFilter;

import org.rosuda.JRI.REXP;

import com.ef_prime.rflow.exception.CanceledException;
import com.ef_prime.rflow.exception.ReadException;
import com.ef_prime.rflow.flow.ViewLayout;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.io.ElementReader;
import com.ef_prime.rflow.io.ElementWriter;
import com.ef_prime.rflow.io.ElementWriterFactory;
import com.ef_prime.rflow.io.FlowDataReader;
import com.ef_prime.rflow.io.FlowDataWriter;
import com.ef_prime.rflow.io.GraphReader;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelManager;
import com.ef_prime.rflow.node.base.FreeNodeModel;
import com.ef_prime.rflow.node.base.SentenceNodeModel;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.tool.AboutDialog;
import com.ef_prime.rflow.tool.DirectoryChooser;
import com.ef_prime.rflow.tool.FileSettingDialog;
import com.ef_prime.rflow.tool.RflowFileFilterFactory;
import com.ef_prime.rflow.tool.SearchDialog;
import com.ef_prime.rflow.util.AbstractActionManager;
import com.ef_prime.rflow.util.FileChooserUtility;
import com.ef_prime.rflow.util.PropertiesUtility;
import com.ef_prime.rflow.xml.XMLException;

public class ActionManager extends AbstractActionManager<ActionManager.ActionKey> implements ItemListener{
	
	private final Rflow rflow;
	
	public static enum ActionKey{
		NEW,
		OPEN,
		SAVE,
		SAVE_AS,
		FILE_OPTION,
		IMPORT_SOURCE,
		EXIT,
		UNDO,
		REDO,
		CUT,
		COPY,
		PASTE,
		FIND,
		COPY_SCRIPT,
		CONSOLE_WINDOW,
		//CODE_VIEWER,
		NEW_NODE,
		NEW_FREE_NODE,
		EDIT_NODE,
		REMOVE_NODE,
		SELECT_ALL,
		RUN,
		RUN_CLEAR,
		STOP,
		CHANGE_DIRECTORY,
		PREFERENCES,
		HELP,
		ABOUT,
		TEST_ACTION;
		
		public static final String SELECTED_KEY = "selected_key"; //$NON-NLS-1$
		
	}
	
	ActionManager(final Rflow rFlow){
		this.rflow = rFlow;
		
		final int mask = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
		
		putAction(ActionKey.NEW, new AbstractAction(Messages.getString("Action.new")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				if(!askSave())
					return;
				rflow.init();
			}
		});
		getAction(ActionKey.NEW).putValue(Action.SMALL_ICON, FileResource.ACTION_NEW);
		getAction(ActionKey.NEW).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_N, mask));

		
		putAction(ActionKey.OPEN, new AbstractAction(Messages.getString("Action.open")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				if(!askSave())
					return;
				
				FileFilter[] filters = new FileFilter[]{RflowFileFilterFactory.createRflowFileFilter()};
				
				File file = FileChooserUtility.chooseOpenFile(Messages.getString("Action.open"), Prefs.WORKING_DIRECTORY, filters, rflow.frame); //$NON-NLS-1$
				if(file == null)return;
				
				open(file, true);
			}
		});
		getAction(ActionKey.OPEN).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_O, mask));
		
		
		putAction(ActionKey.SAVE, new AbstractAction(Messages.getString("Action.save")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				save();
			}
		});
		getAction(ActionKey.SAVE).putValue(Action.SMALL_ICON, FileResource.ACTION_SAVE);
		getAction(ActionKey.SAVE).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_S, mask));
		
		
		putAction(ActionKey.SAVE_AS, new AbstractAction(Messages.getString("Action.save_as")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				save_as();
			}
		});

		putAction(ActionKey.FILE_OPTION, new AbstractAction(Messages.getString("Action.file_option")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				FlowSetting setting = FileSettingDialog.showSettingDialog(rflow.frame, rflow.getFileSetting());
				if(setting != null)
					rflow.setFileSetting(setting);
			}
		});

		putAction(ActionKey.IMPORT_SOURCE, new AbstractAction(Messages.getString("Action.import.source")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				if(!askSave())
					return;
				
				ViewLayout layout = RflowUtil.importSource(rflow.frame);
				if(layout == null)return;
				
				rflow.init();
				rflow.viewer.addSubLayout(layout, null);
			}
		});

		putAction(ActionKey.EXIT, new AbstractAction(Messages.getString("Action.exit")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				if(!askSave())
					return;
				
				System.exit(0);
			}
		});
		
		putAction(ActionKey.UNDO, new AbstractAction(Messages.getString("Action.undo")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.undo();
			}
		});
		getAction(ActionKey.UNDO).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_Z, mask));

		
		putAction(ActionKey.REDO, new AbstractAction(Messages.getString("Action.redo")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.redo();
			}
		});
		getAction(ActionKey.REDO).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_Y, mask));
		
		
		putAction(ActionKey.CUT, new AbstractAction(Messages.getString("Action.cut")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.cut();
			}
		});
		getAction(ActionKey.CUT).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_X, mask));
		
		
		putAction(ActionKey.COPY, new AbstractAction(Messages.getString("Action.copy")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.copy();
			}
		});
		getAction(ActionKey.COPY).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_C, mask));
		
		
		putAction(ActionKey.PASTE, new AbstractAction(Messages.getString("Action.paste")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.paste();
			}
		});
		getAction(ActionKey.PASTE).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_V, mask));
		
		
		putAction(ActionKey.FIND, new AbstractAction(Messages.getString("Action.find")){ //$NON-NLS-1$
			
			private SearchDialog dialog;
			
			public void actionPerformed(ActionEvent e) {
				if(dialog == null){
					dialog = new SearchDialog(rflow.frame, rflow);
					dialog.setLocationRelativeTo(rflow.frame);
				}
				dialog.setVisible(true);
			}
		});
		getAction(ActionKey.FIND).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_F, mask));
		
		
		putAction(ActionKey.COPY_SCRIPT, new AbstractAction(Messages.getString("Action.copy_script")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.copyScript(rflow.state.getSelectedNode());
			}
		});
		
		putAction(ActionKey.CONSOLE_WINDOW, new AbstractAction("コンソール表示"){
			public void actionPerformed(ActionEvent e) {
				rflow.console.setVisible(!rflow.console.isVisible());
				//getAction(Actions.CONSOLE_WINDOW).putValue(Action.SELECTED_KEY, rflow.console.isVisible());
			}
		});
		getAction(ActionKey.CONSOLE_WINDOW).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_F9, 0));
		getAction(ActionKey.CONSOLE_WINDOW).addPropertyChangeListener(new PropertyChangeListener(){
			public void propertyChange(PropertyChangeEvent evt) {
				if(evt.getPropertyName().equals(ActionKey.SELECTED_KEY)){
					if(((Boolean)evt.getNewValue()).booleanValue())
						getAction(ActionKey.CONSOLE_WINDOW).putValue(Action.SMALL_ICON, FileResource.ACTION_REC);
					else
						getAction(ActionKey.CONSOLE_WINDOW).putValue(Action.SMALL_ICON, null);
				}					
			}
		});

		/*putAction(ActionKey.CODE_VIEWER, new AbstractAction("コードビュー"){
			public void actionPerformed(ActionEvent e) {
				rflow.codeViewer.getDialog().setVisible(true);
			}
		});*/
		
		putAction(ActionKey.NEW_NODE, new AbstractAction(Messages.getString("Action.new_sentence_node")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				try {
					NodeModel model = NodeModelManager.getModelController(SentenceNodeModel.class).createNewModel();
					rflow.viewer.addNode(new FlowNode(model), null, true);
				} catch (CanceledException e1) {}
			}
		});
		getAction(ActionKey.NEW_NODE).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, mask));

		
		putAction(ActionKey.NEW_FREE_NODE, new AbstractAction(Messages.getString("Action.new_free_node")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				try {
					NodeModel model = NodeModelManager.getModelController(FreeNodeModel.class).createNewModel();
					rflow.viewer.addNode(new FlowNode(model), null, true);
				} catch (CanceledException e1) {}
			}
		});
		
		
		putAction(ActionKey.EDIT_NODE, new AbstractAction(Messages.getString("Action.edit_node")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				FlowNode node = rflow.state.getSelectedNode();
				if(node == null)return;
				try {
					node.setModel(node.getModel().showConfigDialog(rflow.frame));
				} catch (CanceledException e1) {}
				
			}
		});
		getAction(ActionKey.EDIT_NODE).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0));
		
		
		putAction(ActionKey.REMOVE_NODE, new AbstractAction(Messages.getString("Action.remove_node")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.delete();
			}
		});
		getAction(ActionKey.REMOVE_NODE).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0));
		
		
		putAction(ActionKey.SELECT_ALL, new AbstractAction(Messages.getString("Action.select_all")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.selectAll();
			}
		});
		getAction(ActionKey.SELECT_ALL).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_A, mask));
		
		
		putAction(ActionKey.RUN, new AbstractAction(Messages.getString("Action.run")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.executeRest(rflow.state.getSelectedNode());
			}
		});
		getAction(ActionKey.RUN).putValue(Action.SMALL_ICON, FileResource.ACTION_RUN);
		getAction(ActionKey.RUN).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_R, mask));
		

		putAction(ActionKey.RUN_CLEAR, new AbstractAction(Messages.getString("Action.run_clear")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				rflow.viewer.execute(rflow.state.getSelectedNode(), true);
			}
		});
		//run_clear.putValue(Action.SMALL_ICON, new ImageIcon("img/action/run.png"));
		getAction(ActionKey.RUN_CLEAR).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_R, InputEvent.ALT_MASK));
		
		
		putAction(ActionKey.STOP, new AbstractAction(Messages.getString("Action.stop")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
    			//TODO windows以外のOSでの動作の確認が必要
				rflow.R.rniStop(0);
				rflow.receiver.passCommand(Constants.NEWLINE_STR);
			}
		});
		getAction(ActionKey.STOP).putValue(Action.SMALL_ICON, FileResource.ACTION_STOP);
		getAction(ActionKey.STOP).putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0));
		
		/*
		putAction(Actions.REC, new AbstractAction("記録"){
			public void actionPerformed(ActionEvent e) {
			}
		});
		getAction(Actions.REC).putValue(Action.SMALL_ICON, ImageResource.ACTION_REC);
		*/
	
		putAction(ActionKey.CHANGE_DIRECTORY, new AbstractAction(Messages.getString("Action.change_directory")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				changeDirectory();
			}
		});

		putAction(ActionKey.PREFERENCES, new AbstractAction(Messages.getString("Action.preferences")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				Setting setting = SettingDialog.showSettingDialog(rflow.frame, Rflow.setting, false);
				if(setting != null){
					try {
						PropertiesUtility.store(setting, FileResource.SETTING_FILE);
					} catch (Exception e1) {
						JOptionPane.showMessageDialog(rflow.frame, Messages.getString("Error.write_option"), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
						return;
					}
					Rflow.setting.setPropertyAll(setting);
				}
			}
		});
		
		putAction(ActionKey.HELP, new AbstractAction(Messages.getString("Action.help")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
		  		//OSの確認
				switch(Platform.getPlatform()){
				case WINDOWS:
				case WINDOWS_VISTA:
			  		if(!FileResource.CHTML_HELP_FILE.exists()){
			  			JOptionPane.showMessageDialog(rflow.frame, Messages.getString("Error.no_help")); //$NON-NLS-1$
			  			return;
			  		}
			  		try {
						Runtime.getRuntime().exec("hh.exe "+FileResource.CHTML_HELP_FILE.getAbsolutePath()); //$NON-NLS-1$
					} catch (IOException ex) {
			  			JOptionPane.showMessageDialog(rflow.frame, Messages.getString("Error.run_help")); //$NON-NLS-1$
						ex.printStackTrace();
						return;
					}
					break;
				default://not windows
					//javahelp
					if(FileResource.JHELP_FILE == null || !FileResource.JHELP_FILE.exists()){
						JOptionPane.showMessageDialog(rflow.frame, Messages.getString("Error.no_help")); //$NON-NLS-1$
						return;
					}
					try{
						URL url = FileResource.JHELP_FILE.toURL();
						HelpSet helpset = new HelpSet(ClassLoader.getSystemClassLoader(), url);
	
						HelpBroker broker = helpset.createHelpBroker();
						ActionListener helper = new CSH.DisplayHelpFromSource(broker);
						helper.actionPerformed(e);
					}catch (Exception e1) {
			  			JOptionPane.showMessageDialog(rflow.frame, Messages.getString("Error.run_help")); //$NON-NLS-1$
			  			if(Rflow.DEBUG > 0)e1.printStackTrace();
						return;
					}
				}
			}
		});
		
		
		putAction(ActionKey.ABOUT, new AbstractAction(Messages.getString("Action.about")){ //$NON-NLS-1$
			public void actionPerformed(ActionEvent e) {
				AboutDialog dialog = new AboutDialog(rflow.frame);
				dialog.setLocationRelativeTo(rflow.frame);
				dialog.setVisible(true);
			}
		});

		
		//初期不可
		getAction(ActionKey.UNDO).setEnabled(false);
		getAction(ActionKey.REDO).setEnabled(false);
		getAction(ActionKey.CUT).setEnabled(false);
		getAction(ActionKey.COPY).setEnabled(false);
		getAction(ActionKey.PASTE).setEnabled(false);
		getAction(ActionKey.REMOVE_NODE).setEnabled(false);
		getAction(ActionKey.COPY_SCRIPT).setEnabled(false);
		getAction(ActionKey.EDIT_NODE).setEnabled(false);
		getAction(ActionKey.RUN).setEnabled(false);
		getAction(ActionKey.RUN_CLEAR).setEnabled(false);
	}
	
	
	JMenuBar createMenuBar(){
		JMenuBar menuBar = new JMenuBar();
		
		JMenu file = new JMenu(Messages.getString("Menu.file")); //$NON-NLS-1$
		file.add(getAction(ActionKey.NEW));
		file.add(getAction(ActionKey.OPEN));
		file.addSeparator();
		file.add(getAction(ActionKey.SAVE));
		file.add(getAction(ActionKey.SAVE_AS));
		file.addSeparator();
		file.add(getAction(ActionKey.IMPORT_SOURCE));
		file.addSeparator();
		file.add(getAction(ActionKey.EXIT));

		JMenu edit = new JMenu(Messages.getString("Menu.edit")); //$NON-NLS-1$
		edit.add(getAction(ActionKey.UNDO));
		edit.add(getAction(ActionKey.REDO));
		edit.addSeparator();
		edit.add(getAction(ActionKey.CUT));
		edit.add(getAction(ActionKey.COPY));
		edit.add(getAction(ActionKey.PASTE));
		edit.addSeparator();
		edit.add(getAction(ActionKey.REMOVE_NODE));
		edit.add(getAction(ActionKey.SELECT_ALL));
		edit.addSeparator();
		edit.add(getAction(ActionKey.FIND));
		edit.addSeparator();
		edit.add(getAction(ActionKey.COPY_SCRIPT));
		
//		JMenu display = new JMenu("表示");
//		display.add(getAction(Actions.CONSOLE_WINDOW));

		JMenu node = new JMenu(Messages.getString("Menu.node")); //$NON-NLS-1$
		node.add(getAction(ActionKey.NEW_NODE));
		node.add(getAction(ActionKey.NEW_FREE_NODE));
		node.add(getAction(ActionKey.EDIT_NODE));
		
		JMenu exec = new JMenu(Messages.getString("Menu.run")); //$NON-NLS-1$
		exec.add(getAction(ActionKey.RUN));
		exec.add(getAction(ActionKey.RUN_CLEAR));
		exec.add(getAction(ActionKey.STOP));

		JMenu option = new JMenu(Messages.getString("Menu.option")); //$NON-NLS-1$
		option.add(getAction(ActionKey.CHANGE_DIRECTORY));
		option.add(getAction(ActionKey.FILE_OPTION));
		option.addSeparator();
		option.add(getAction(ActionKey.PREFERENCES));

		JMenu help = new JMenu(Messages.getString("Menu.help")); //$NON-NLS-1$
		help.add(getAction(ActionKey.HELP));
		help.add(getAction(ActionKey.ABOUT));
		//if(Rflow.TEST)help.add(getAction(ActionKey.CODE_VIEWER));
		//if(Rflow.TEST)help.add(getAction(ActionKey.TEST_ACTION));
		

		menuBar.add(file);
		menuBar.add(edit);
//		menuBar.add(display);
		menuBar.add(node);
		menuBar.add(exec);
		menuBar.add(option);
		menuBar.add(help);
		
		
		return menuBar;
	}
	
	JToolBar createToolBar(){
		JToolBar toolBar = new JToolBar();

		toolBar.add(getAction(ActionKey.NEW));
		toolBar.add(getAction(ActionKey.SAVE));
		toolBar.addSeparator();
		toolBar.add(getAction(ActionKey.RUN));
		toolBar.add(getAction(ActionKey.STOP));
		//toolBar.add(getAction(Actions.REC));//test
		
		return toolBar;
	}
	
	public void itemStateChanged(ItemEvent e) {
		boolean isSelection = (rflow.state.getPickedVertices().size() > 0);
		getAction(ActionKey.CUT).setEnabled(isSelection);
		getAction(ActionKey.COPY).setEnabled(isSelection);
		getAction(ActionKey.REMOVE_NODE).setEnabled(isSelection);
		
		boolean singleSelection = (rflow.state.getPickedVertices().size() == 1);
		getAction(ActionKey.COPY_SCRIPT).setEnabled(singleSelection);
		getAction(ActionKey.EDIT_NODE).setEnabled(singleSelection);
		getAction(ActionKey.RUN).setEnabled(singleSelection);
		getAction(ActionKey.RUN_CLEAR).setEnabled(singleSelection);
	}
	
	/**
	 * .rflowファイルを開く
	 * @param file .rflowファイル
	 */
	public void open(File file, boolean clearObject){
		if(file == null)return;
		try {
			FlowDataReader reader = new FlowDataReader();
			GraphReader graphReader = new GraphReader(rflow.frame, rflow.viewer.getSize());
			//FunctionTreeReader functionTreeReader = new FunctionTreeReader();
			//PartsTreeReader partsTreeReader = new PartsTreeReader();
			
			reader.read(file, new ElementReader[]{
					graphReader, 
					//functionTreeReader, 
					//partsTreeReader
					});
			
			//初期化
			rflow.init(clearObject, file, graphReader.getFileSettings(), file.getParentFile());
			rflow.viewer.init(graphReader.getViewLayout());
			//rflow.explorer.getFunctionExplorerPanel().init(functionTreeReader.getRootNode());
			//rflow.explorer.getPartsExplorerPanel().init(partsTreeReader.getRoot());

			
		} catch (ReadException e1) {
			if(Rflow.DEBUG > 0)e1.printStackTrace();
			JOptionPane.showMessageDialog(rflow.frame, e1.getMessage(), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
			return;
		} catch (IOException e1) {
			if(Rflow.DEBUG > 0)e1.printStackTrace();
			JOptionPane.showMessageDialog(rflow.frame, e1.getMessage(), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
			return;
		} catch (Exception e1){
			if(Rflow.DEBUG > 0)e1.printStackTrace();
			JOptionPane.showMessageDialog(rflow.frame, Messages.getString("Error.read"), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
			return;
		}
	}
	
	/**
	 * 上書き保存
	 * @return　false if failed or save canceled
	 */
	private boolean save(){
		File file = rflow.getWorkingFile();
		if(file == null)
			return save_as();
		else
			return save_internal(file);
	}
	
	/**
	 * 名前を付けて保存
	 * @return false if failed or save canceled
	 */
	private boolean save_as(){
		FileFilter[] filters = new FileFilter[]{RflowFileFilterFactory.createRflowFileFilter()};
		
		File file = FileChooserUtility.chooseSaveFile(Messages.getString("Message.save"), Prefs.WORKING_DIRECTORY, filters, rflow.frame); //$NON-NLS-1$
		return save_internal(file);
	}
	
	private boolean save_internal(File file){
		if(file == null)return false;
		try {
			ElementWriter[] elementWriters = new ElementWriter[]{
					ElementWriterFactory.getGraphWriter(rflow.viewer.getViewLayout(), rflow.getFileSetting()),
					//ElementWriterFactory.getFunctionTreeElementWriter(rflow.explorer.getFunctionExplorerPanel().getRoot()),
					//ElementWriterFactory.getPartsTreeWriter(rflow.explorer.getPartsExplorerPanel().getRoot())
					};
			new FlowDataWriter().write(file, elementWriters);
			
			rflow.setWorkingFile(file);
			rflow.setWorkingDirectory(file.getParentFile());
			rflow.setDirty(false);
			
			return true;
		} catch (IOException e1) {
			JOptionPane.showMessageDialog(rflow.frame, e1.getMessage(), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$
			return false;
		} catch (XMLException e) {
			JOptionPane.showMessageDialog(rflow.frame, Messages.getString("Error.save"), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE); //$NON-NLS-1$ //$NON-NLS-2$
			return false;
		} 
	}
	
	/**
	 * 保存の確認
	 * @return false if canceled or save failed
	 */
	public boolean askSave(){
		if(!rflow.isDirty())return true;
		
		int ret = JOptionPane.showConfirmDialog(rflow.frame, Messages.getString("Message.confirm_save"), Messages.getString("Message.confirm"), JOptionPane.YES_NO_CANCEL_OPTION); //$NON-NLS-1$ //$NON-NLS-2$
		
		switch(ret){
		case JOptionPane.YES_OPTION:
			return save();
		case JOptionPane.NO_OPTION:
			return true;
		default:
			return false;
		}
	}


	void changeDirectory() {
		File currentDir = Prefs.WORKING_DIRECTORY; //$NON-NLS-1$
		REXP wdEXP = RCommand.getR().eval("try(getwd(), silent=TRUE)"); //$NON-NLS-1$
		
		if(wdEXP != null)
			currentDir = new File(Constants.replaceRSep2SystemSep(wdEXP.asString()));				
		
		File ret = DirectoryChooser.chooseDirectory(rflow.frame, Messages.getString("Action.change_directory"), currentDir); //$NON-NLS-1$
		
		if(ret != null){
			rflow.setWorkingDirectory(ret);
		}
	}
}
