package com.ef_prime.rflow;

import java.io.File;
import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Vector;

import org.rosuda.JRI.RBool;
import org.rosuda.JRI.REXP;
import org.rosuda.JRI.RList;
import org.rosuda.JRI.RVector;
import org.rosuda.JRI.Rengine;

import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.base.ArgEntry;
import com.ef_prime.rflow.node.base.FreeNodeModel;
import com.ef_prime.rflow.node.base.SentenceNodeModel;

public class RCommand {
	
	private static Rengine R;
	private static CommandReceiver receiver;

	public static void setR(Rengine R) {
		RCommand.R = R;
	}
	
	public static void setReceiver(CommandReceiver receiver) {
		RCommand.receiver = receiver;
	}
	
	public static Rengine getR(){
		return R;
	}
	
	public static CommandReceiver getReceiver() {
		return receiver;
	}
	
	public static String[] supportList(String s){
		if(!checkCommandString(s))return null;
		
		REXP cmds = R.eval("try(.supportCommand(\""+s+"\"), silent = TRUE)"); //$NON-NLS-1$ //$NON-NLS-2$
		String[] c = null;
        if (cmds != null && (c=cmds.asStringArray()) != null) return c;
		return null;
	}
	
	public static String[] supportListGlobal(String s){
		if(s == null)
			s = ""; //$NON-NLS-1$
		
		REXP cmds = R.eval("try(.supportCommand(\""+s+"\", envir=.GlobalEnv), silent = TRUE)"); //$NON-NLS-1$ //$NON-NLS-2$
		String[] c = null;
        if (cmds != null && (c=cmds.asStringArray()) != null) return c;
		return null;
	}	
	
	public static Map<String, String> argList(String function){
		if(!checkCommandString(function))return null;
		
		REXP formals = R.eval("try(.getFormals(\""+function+"\"), silent = TRUE)"); //$NON-NLS-1$ //$NON-NLS-2$
		if(formals == null)return null;
		RList list = formals.asList();
		if(list == null)return null;
		
		Map<String, String> map = new LinkedHashMap<String, String>();
		String[] keys = list.keys();
		for(int i = 0; i < keys.length; i++){
			String key = keys[i];
			String value = list.at(i).asString();
			//System.out.println(key + ":" + value);
			map.put(key, value);
		}
		
		return map;
	}
	
	public static String[] getAllVars(String expression){
		if(!checkCommandString(expression))return null;
		
		REXP vars = R.eval("all.vars(expression("+expression+"))"); //$NON-NLS-1$ //$NON-NLS-2$
		if(vars == null)return null;
		return vars.asStringArray();		
	}
	
	public static String[] getLocalVariables(String function){
		if(!checkCommandString(function))return null;
		
		REXP locals = R.eval(".findLocals("+function+")"); //$NON-NLS-1$ //$NON-NLS-2$
		if(locals == null)return null;
		RList list = locals.asList();
		if(list == null)return null;
		
		REXP variables = list.at("variables"); //$NON-NLS-1$
		String[] c = null;
        if (variables != null && (c=variables.asStringArray()) != null) return c;
		
		return null;
	}
	
	public static void clearObjects(){
		R.eval(".clearObjects()"); //$NON-NLS-1$
	}
	
	public static final String FILTER_DATAFRAME = "is.data.frame"; //$NON-NLS-1$
	public static String[] getObjectsFiltered(String filter){
		REXP objects = R.eval(".getObjectsFiltered(FILTERFUN="+filter+")"); //$NON-NLS-1$ //$NON-NLS-2$
		if(objects == null)return null;
		return objects.asStringArray();
	}

	private static final String localEnv = ".localEnv"; //$NON-NLS-1$
	public static void clearLocalEnv(){
		R.eval("if(exists(\""+localEnv+"\"))rm("+localEnv+")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	}
	
	public static String getLocalExcutionCode(String command){
		if(!checkCommandString(command))return null;
		
		return ".doInLocal(parse(text=\""+command+"\"), envir="+localEnv+", envchar=\""+ localEnv+"\")\n"; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
	}
	
	public static void help(final String function){
		R.eval("help(\""+function+"\")"); //$NON-NLS-1$ //$NON-NLS-2$
	}
	
	
	/*
	 * コードからフローのモデルを作成する
	 */
	private static final String KEY_LINE = "line"; //$NON-NLS-1$
	private static final String KEY_FUNC = "func"; //$NON-NLS-1$
	private static final String KEY_ARGS = "args"; //$NON-NLS-1$
	private static final String KEY_SYMBOL = "symbol"; //$NON-NLS-1$
	private static final String KEY_SENTENCE = "sentence"; //$NON-NLS-1$
	private static final String KEY_CONTROL = "control"; //$NON-NLS-1$
	
	public static NodeModel[] parseToNodes(File file, boolean flowControlIsFreeNode){
		return parseToNodesInternal(null, file, flowControlIsFreeNode);
	}

	public static NodeModel[] parseToNodes(String code, boolean flowControlIsFreeNode){
		return parseToNodesInternal(code, null, flowControlIsFreeNode);
	}
	
	private static NodeModel[] parseToNodesInternal(String code, File file, boolean flowControlIsFreeNode){
		REXP codeInfos;
		
		if(code != null){
			codeInfos = R.eval("try(parseToNodes(\""+ escapeDoubleQuote(code) +"\"), silent=TRUE)"); //$NON-NLS-1$ //$NON-NLS-2$
		}else if(file != null){
			String filePath;
			try {
				filePath = Constants.replaceFileSeparator(file.getCanonicalPath());
			} catch (IOException e) {
				return null;
			} 
			codeInfos = R.eval("try(parseToNodes(file=\""+ escapeDoubleQuote(filePath) +"\"), silent=TRUE)"); //$NON-NLS-1$ //$NON-NLS-2$
		}else{
			return null;
		}
		
		
		if(codeInfos == null)return null;
		
		RVector codeInfoV = codeInfos.asVector();
		if(codeInfoV == null)return null;
		//RList codeInfoList = codeInfos.asList();
		//if(codeInfoList == null)return null;
		
		int nLine = codeInfoV.size();
		
		NodeModel[] models = new NodeModel[nLine];
		boolean[] isFreeNode = new boolean[models.length];
		
		for(int i = 0; i < nLine; i++){
			REXP codeInfo = codeInfoV.at(i);
			NodeModel model;
			
			if(!(codeInfo.getType() == REXP.XT_VECTOR || codeInfo.getType() == REXP.XT_LIST))
				return null;
			
			RList funcInfo = codeInfo.asList();

			REXP line_exp = funcInfo.at(KEY_LINE);
			if(line_exp == null)return null;
			String line = line_exp.asString();
			
			REXP sentence_exp = funcInfo.at(KEY_SENTENCE);
			if(sentence_exp == null)return null;
			String sentence = sentence_exp.asString();
			boolean isSentence = true;
			isFreeNode[i] = !isSentence;
			
			if(isSentence){
				boolean isControl = false;
				if(flowControlIsFreeNode){
					REXP control = funcInfo.at(KEY_CONTROL);
					if(control != null){
						RBool b =  control.asBool();
						isControl = (b != null && b.isTRUE());
					}
				}
				
				if(isControl){
					FreeNodeModel fModel = new FreeNodeModel();
					fModel.command = line;
					model = fModel;
					isFreeNode[i] = true;
				}else{
					SentenceNodeModel sModel = new SentenceNodeModel();
					
					sModel.sentence = sentence;
	
					REXP function = funcInfo.at(KEY_FUNC);
					if(function != null){
						sModel.function = funcInfo.at(KEY_FUNC).asString();
					}
					
					REXP symbol = funcInfo.at(KEY_SYMBOL);
					if(symbol != null){
						sModel.symbol = symbol.asString();
					}
					
					REXP args = funcInfo.at(KEY_ARGS);
					if(args != null){
						RVector argsList = args.asVector();
						if(argsList != null){
							Vector names = argsList.getNames();
							boolean hasName = (names != null && names.size() > 0);
							for(int j = 0; j < argsList.size(); j++){
								String key = (hasName)?names.get(j).toString():null;
								String v = argsList.at(j).asString();
								sModel.argList.add(new ArgEntry(key, v.trim()));
							}
						}
					}
					
					model = sModel;
				}
			}else{
				//unreachable now
				//free node if not list(one character)
				FreeNodeModel fModel = new FreeNodeModel();
				fModel.command = line;
				model = fModel;
			}

			
			models[i] = model;
		}
		
		//連続するFreeNodeModelを結合する
		LinkedList<NodeModel> modelList = new LinkedList<NodeModel>();
		modelList.add(models[0]);
		for(int i = 1; i< models.length; i++){
			if(isFreeNode[i] && isFreeNode[i-1]){
				FreeNodeModel lastFreeModel = (FreeNodeModel)modelList.getLast();
				FreeNodeModel currentFreeModel = (FreeNodeModel)models[i];
				
				String c = lastFreeModel.getCommandNewline();
				FreeNodeModel fm = new FreeNodeModel();
				fm.command = c + currentFreeModel.command;
				
				modelList.removeLast();
				modelList.add(fm);
			}else{
				modelList.add(models[i]);
			}
		}
		
		return modelList.toArray(new NodeModel[0]);
	}

	
	private static boolean checkCommandString(String command){
		if(command == null || (command.trim().length() < 1))
			return false;
		return true;		
	}
	
	private static String escapeDoubleQuote(String s){
		return s.replaceAll("\"", "\\\\\""); //$NON-NLS-1$ //$NON-NLS-2$
	}
}
