package com.ef_prime.rflow;

import java.awt.BorderLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Locale;

import javax.swing.Box;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JToolBar;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import org.rosuda.JRI.REXP;
import org.rosuda.JRI.Rengine;

import com.ef_prime.rflow.Setting.Key;
import com.ef_prime.rflow.console.RConsole;
import com.ef_prime.rflow.explorer.Explorer;
import com.ef_prime.rflow.flow.FlowState;
import com.ef_prime.rflow.flow.FlowViewer;
import com.ef_prime.rflow.flow.FlowViewerScrollPane;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.group.AppearanceGroupManager;
import com.ef_prime.rflow.node.DefaultNodeModelController;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.node.NodeModelChangeListener;
import com.ef_prime.rflow.node.NodeModelController;
import com.ef_prime.rflow.node.NodeModelManager;
import com.ef_prime.rflow.node.base.FreeNodeModel;
import com.ef_prime.rflow.node.base.FreeNodeModelController;
import com.ef_prime.rflow.node.base.SentenceNodeModel;
import com.ef_prime.rflow.node.base.SentenceNodeModelController;
import com.ef_prime.rflow.node.base.SubflowNodeModel;
import com.ef_prime.rflow.node.base.SubflowNodeModelController;
import com.ef_prime.rflow.node.base.TunnelNodeModel;
import com.ef_prime.rflow.node.base.TunnelNodeModelController;
import com.ef_prime.rflow.resource.Messages;
import com.ef_prime.rflow.tool.UpdateChecker;
import com.ef_prime.rflow.tool.WorkingDirSetter;
import com.ef_prime.rflow.util.FileControl;
import com.ef_prime.rflow.util.PropertiesUtility;
import com.ef_prime.rflow.xml.XMLException;

import edu.uci.ics.jung.graph.event.GraphEventType;

public class Rflow {
	
	/**
	 * デバッグ用定数
	 * この値はそのままRengine.DEBUGにも代入されます
	 * リリース時には0に設定する
	 * <p>0: デバッグ出力なし
	 * <br>>0:デバッグ出力あり
	 * <p>例：
	 * <code>if(Rflow.DEBUG > 0)
	 *     System.out.println(e.getMessage());</code>
	 */
	public static final int DEBUG = 0;
	
	/**
	 * プログラムのバージョン
	 */
	public static final String VERSION = "0.3.1"; //$NON-NLS-1$
	
	public final MainFrame frame;
	public final Rengine R;
	public final CommandReceiver receiver;
	public final FlowState state;
	public final FlowViewer viewer;
	public final CodeBar codeBar;
	public final CodeViewer codeViewer;
	public final Explorer explorer;
	public final RConsole console;
	public final ActionManager actionManager;
	public final AppearanceGroupManager groupManager = new AppearanceGroupManager();
	public static final Setting setting = new Setting();
	
	private File workingFile;
	private FlowSetting fileSetting = new FlowSetting();
	private boolean isDirty;
	
	static{
		Rengine.DEBUG = Rflow.DEBUG;	//set debug
	}
	
	public Rflow(){		
		/*SplashScreen splashScreen = SplashScreen.getSplashScreen();
		try {
			splashScreen.setImageURL(new File("img/R-flow-splash.png").toURI().toURL());
		}　catch (Exception e) {
			e.printStackTrace();
		}*/
		
		//System.getProperties().list(System.out);
		//System.out.println(System.getProperty("user.dir"));
		
		//Look&Feel
		setLookAndFeel();

		//設定ファイルの読み込み
		try {
			PropertiesUtility.load(setting, FileResource.SETTING_FILE);
			setting.update();
		} catch (IOException e2) {
			e2.printStackTrace();
			JOptionPane.showMessageDialog(null, Messages.getString("Error.read_option"), Messages.getString("Error.title.error"), JOptionPane.ERROR_MESSAGE);  //$NON-NLS-1$//$NON-NLS-2$
		}

		//ロケールの設定
		Locale.setDefault(new Locale(((String)setting.getProperty(Key.LANGUAGE))));

		//各種ファイルを設定をロケールにしたがって設定
		FileResource.prepareFiles();
		
		//リソースの設定（設定ファイルの言語設定を使う）
		setResource();

		
		//メインフレーム
		frame = new MainFrame();
		frame.addWindowListener(new WindowAdapter(){
			@Override
			public void windowClosing(WindowEvent e) {
				if(!actionManager.askSave()){
					return;
				}
				
				onExit();
				
				System.exit(0);
			}
		});
		
		//ノードモデル
		initializeNodeModelManager();
		
		//パネル
		JPanel panel = new JPanel(new BorderLayout());
		//レシーバー
		receiver = new CommandReceiver(this);
		//フロー
		Box flowBox = Box.createVerticalBox();
		//ステート
		state = new FlowState();
		//コードバー
		codeBar = new CodeBar();
		state.addItemListener(codeBar.pickedStateItemListener);
		flowBox.add(codeBar);
		//ビュー
		viewer = new FlowViewer(this, state);
		//GraphZoomScrollPane graphPane = new GraphZoomScrollPane(viewer);
		FlowViewerScrollPane graphPane = new FlowViewerScrollPane(viewer);
		flowBox.add(graphPane);
		//関数
		explorer = new Explorer(this);
		//中央スプリットペイン
		JSplitPane centerSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, explorer, flowBox);
		centerSplitPane.setDividerLocation(200);
		panel.add(centerSplitPane, BorderLayout.CENTER);
		//アクション
		actionManager = new ActionManager(this);
		state.addItemListener(actionManager);
		//ツールバー
		JToolBar toolbar = actionManager.createToolBar();
		panel.add(toolbar, BorderLayout.NORTH);
		//メニューバー
		JMenuBar menuBar = actionManager.createMenuBar();
		
		frame.add(panel, BorderLayout.CENTER);
		frame.add(menuBar, BorderLayout.NORTH);
		//frame.setLocationRelativeTo(null);

		//コードビューワ
		codeViewer = new CodeViewer(this);
		state.addItemListener(codeViewer);
		viewer.getGraph().addListener(codeViewer, GraphEventType.ALL_SINGLE_EVENTS);
		
		
		//node model change listener
		FlowNode.addNodeModelChangeListener(state);
		FlowNode.addNodeModelChangeListener(codeBar.modelChangeListener);
		FlowNode.addNodeModelChangeListener(viewer);
		FlowNode.addNodeModelChangeListener(codeViewer);
		FlowNode.addNodeModelChangeListener(new NodeModelChangeListener(){
			public void nodeModelChanged(FlowNode node, NodeModel old, NodeModel model) {
				setDirty(true);
			}
		});
		
		
		//R console
		console = new RConsole(this);
		
		//R
		R = new Rengine(new String[0], false, console);
		if(!R.waitForR()){
			return;
		}	
		
		//
		RCommand.setR(R);
		RCommand.setReceiver(receiver);
		
		//警告表示の抑制
		//R.eval("old.warn <- getOption(\"warn\")");
		//R.eval("options(warn=-1)");


		//デフォルトの作業ディレクトリ
		String dir = (String)setting.getProperty(Setting.Key.DEFAULT_WORKING_DIRECTORY);
		if(!WorkingDirSetter.setWorkingDir(
				new FileControl((Integer)Rflow.setting.getProperty(Setting.Key.MAX_PATH), frame)
				, frame, (dir.equals(""))?null:new File(dir))) //$NON-NLS-1$
			System.exit(0);

		//init.Rが存在したらその内容を利用する.
		File initR = new File(Platform.getUserDir(),"init.R"); //$NON-NLS-1$
		
		if(initR.exists())
		{
			//Rのコマンド行う場合：ユーザ名の文字コードとの関係がわからないのコメントアウト
			//R.eval("source(\""+ Constants.replaceFileSeparator(initR.getAbsolutePath()) +"\")");
			BufferedReader reader = null;
			try {
				reader = new BufferedReader(new FileReader(initR));

				String line;
				while((line = reader.readLine()) != null)
					R.eval(line);
				
			} catch (FileNotFoundException e1) {
			} catch (IOException e) {
			} finally{
				try {
					if(reader != null)
						reader.close();
				} catch (IOException e1) {}
			}
		}
		else
		{
			R.eval("library(\"rJava\")"); //$NON-NLS-1$
			R.eval("library(\"JavaGD\")"); //$NON-NLS-1$
		}

		R.eval("source(\"initRflow.R\", encoding=\"UTF-8\")"); //$NON-NLS-1$
		
		//rflow用の関数群の登録
		R.eval("attach(new.env(hash=TRUE, parent=.GlobalEnv), name=\""+ Constants.RFLOW_ENV +"\")"); //$NON-NLS-1$ //$NON-NLS-2$
		R.eval("eval(quote(source(\"rflow.R\", encoding=\"UTF-8\", local=TRUE)), envir=as.environment(\""+ Constants.RFLOW_ENV +"\"))"); //$NON-NLS-1$ //$NON-NLS-2$
		R.eval("eval(quote(source(\"func.R\", encoding=\"UTF-8\", local=TRUE)), envir=as.environment(\""+ Constants.RFLOW_ENV +"\"))"); //$NON-NLS-1$ //$NON-NLS-2$
		R.eval("lockEnvironment(as.environment(\""+ Constants.RFLOW_ENV +"\"))"); //$NON-NLS-1$ //$NON-NLS-2$
		
		//ツリー初期化
		try {

			File file = FileResource.USER_FUNCTION_FILE;
			if(!file.exists())
				file = FileResource.FUNCTION_FILE;
			
			if(file.exists())
				explorer.getFunctionExplorer().importFunctionTree(file, groupManager);
		} catch (Exception e) {
			//JOptionPane.showMessageDialog(frame, "関数ツリーの読み込みに失敗しました。", "エラー", JOptionPane.ERROR_MESSAGE);
		}
		
		//警告表示を元に戻す
		//R.eval("options(warn=old.warn)");
		//R.eval("rm(old.warn)");
		
		//初期化
		init(false, null, null, new File((String)setting.getProperty(Setting.Key.DEFAULT_WORKING_DIRECTORY)));
		
		//コンソール用のメインループを開始
		R.startMainLoop();
		//receiver.passCommand("Sys.setlocale(\"LC_ALL\", \""+Locale.getDefault().getLanguage()+"\")");
		
		try {
			SwingUtilities.invokeAndWait(new Runnable(){
				public void run() {
					frame.setVisible(true);
					console.setVisible(true);
				}});
		} catch (InterruptedException e1) {
		} catch (InvocationTargetException e1) {
		}

		checkUpdate();
	}
	
	private static void setLookAndFeel(){
		//Linux対応
		new JPanel();
		//look&feel
		try{
	        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
//	        UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");
//	        UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
//	        UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
//	        UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsClassicLookAndFeel");
	    }catch(Exception ex){
	    	ex.printStackTrace();
	    }
	}
	
	private void setResource(){
		//リソースバンドルを言語設定にしたがって取得
		Messages.prepareResourceBundle();
		//Swing Component のリソースを設定
		UIManager.put("AbstractUndoableEdit.undoText", Messages.getString("Action.undo")); //$NON-NLS-1$ //$NON-NLS-2$
		UIManager.put("AbstractUndoableEdit.redoText", Messages.getString("Action.redo")); //$NON-NLS-1$ //$NON-NLS-2$
		//UIManager.put("OptionPane.yesButtonText", Messages.getString("Button.yes"));
		//UIManager.put("OptionPane.noButtonText", Messages.getString("Button.no"));
		UIManager.put("OptionPane.cancelButtonText", Messages.getString("Button.cancel")); //$NON-NLS-1$ //$NON-NLS-2$
		UIManager.put("OptionPane.okButtonText", Messages.getString("Button.ok")); //$NON-NLS-1$ //$NON-NLS-2$
		UIManager.put("FileChooser.cancelButtonText", Messages.getString("Button.cancel")); //$NON-NLS-1$ //$NON-NLS-2$
//		UIManager.put("OptionPane.yesButtonMnemonic", "はい");
//		UIManager.put("OptionPane.noButtonMnemonic", String.valueOf(KeyEvent.VK_N));
//		UIManager.put("OptionPane.cancelButtonMnemonic", "キャンセル");
//		UIManager.put("OptionPane.okButtonMnemonic", "OK");

//		UIManager.getIcon("OptionPane.errorIcon");
//	    UIManager.getIcon("OptionPane.informationIcon");
//	    UIManager.getIcon("OptionPane.warningIcon");
//	    UIManager.getIcon("OptionPane.questionIcon");
	}
	
	private void initializeNodeModelManager(){
		NodeModelManager.addNodeModel(SentenceNodeModel.class, new SentenceNodeModelController());
		NodeModelManager.addNodeModel(FreeNodeModel.class, new FreeNodeModelController());
		NodeModelManager.addNodeModel(SubflowNodeModel.class, new SubflowNodeModelController());
		NodeModelManager.addNodeModel(TunnelNodeModel.class, new TunnelNodeModelController());
		
		for(NodeModelController controller : NodeModelManager.getModelControllerList()){
			if (controller instanceof DefaultNodeModelController) {
				DefaultNodeModelController defCon = (DefaultNodeModelController) controller;
				defCon.setParentWindow(frame);
			}
		}
	}
	
	private void onExit(){
		try {
			explorer.getFunctionExplorer().save(FileResource.USER_FUNCTION_FILE);
			PropertiesUtility.store(setting, FileResource.SETTING_FILE);
		} catch (IOException e) {
			// TODO メッセージの表示を検討
			e.printStackTrace();
		} catch (XMLException e) {
			// TODO メッセージの表示を検討
			e.printStackTrace();
		}
	}


	/**
	 * 新規ファイル時の初期化
	 */
	void init(){
		init(true);
	}
	
	/**
	 * 新規ファイル時の初期化
	 * @param clearObject objectをクリアするか
	 */
	void init(boolean clearObject){
		init(clearObject, null, null, null);
	}
	
	/**
	 * 新規ファイル時の初期化
	 * @param clearObject objectをクリアするか
	 * @param workingFile 作業ファイル
	 * @param fileSetting ファイルの設定
	 * @param workingDirectory 作業ディレクトリ
	 */
	void init(boolean clearObject, 
			File workingFile, 
			FlowSetting fileSetting, 
			File workingDirectory)
	{
		//rflow.R.rniStop(0);
		if(clearObject)
			RCommand.clearObjects();
		viewer.clearFlow();
		if(clearObject)
			console.clearConsole();
		codeBar.clear();
		setWorkingFile(workingFile);
		setFileSetting(fileSetting);
		setWorkingDirectory(workingDirectory);
		setDirty(false);
		//TODO JavaGDを消す
	}
	
	public void setWorkingFile(File workingFile) {
		this.workingFile = workingFile;
		frame.setFileName(workingFile, isDirty());
	}
	
	public File getWorkingFile() {
		return workingFile;
	}
	
	public void setFileSetting(FlowSetting fileSetting) {
		if(fileSetting == null){
			fileSetting = FlowSetting.createDefaultSettings();
		}
		if(!this.fileSetting.equals(fileSetting)){	
			this.fileSetting = fileSetting;
			viewer.updateFileSetting();
			setDirty(true);
		}
	}
	
	public FlowSetting getFileSetting() {
		return fileSetting;
	}
	

	/**
	 * 作業ディレクトリを変更します。
	 * nullを渡すとデフォルト値に設定します。
	 * @param file 
	 */
	public void setWorkingDirectory(File file){
		setWD(file);
		explorer.getFileExplorer().update();
	}
	
	private static void setWD(File file){
		//nullの場合変更しない
		if(file == null){
			return;
		}

		String dir = Constants.replaceFileSeparator(file.getAbsolutePath());
		RCommand.getR().eval("try(setwd(\""+dir+"\"), silent=TRUE)"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-2
		//すぐにgetwd()を実行し変更が正常に終了したかチェックする
		boolean success = false;
		REXP wdEXP = RCommand.getR().eval("try(getwd(), silent=TRUE)"); //$NON-NLS-1$
		if(wdEXP != null){
			success = dir.equals(wdEXP.asString());
		}				
		
		if(success){
			Prefs.WORKING_DIRECTORY = file;
		}
	}
	
	public void setDirty(boolean isDirty) {
		if(this.isDirty != isDirty){
			frame.setFileName(workingFile, isDirty);
		}
		this.isDirty = isDirty;
	}
	
	public boolean isDirty() {
		return isDirty;
	}

	public static void main(String[] args) {	

		//このへんWindows専用かも
		checkIsFirst();
		
		Rflow rflow = 
			new Rflow();
			
		//.rflowファイルを開く
		if(args.length == 1)
		{        	
			File file = new File(args[0]);
			
			if(file.exists())
			{
				rflow.actionManager.open(file, true);
			}
		}
	}


	private static void checkUpdate() {
		try {
			//Setting setting = SerializeUtility.load(settingFile);
			//Setting setting = new Setting();
			//PropertiesUtility.load(setting, settingFile);
			if((Boolean)setting.getProperty(Key.IS_AUTO_UPDATE))
			{
				UpdateChecker.updateCheck();
			}
			
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private static final String NEW_LINE_CODE =
		System.getProperty("line.separator"); //$NON-NLS-1$
	
	private static void checkIsFirst() {

		Setting setting = Setting.defaultSetting();
		try {
			PropertiesUtility.load(setting, FileResource.SETTING_FILE);
		} catch (FileNotFoundException e1) {
			//do nothing
		} catch (IOException e1) {
			//do nothing
		}
		
		Locale.setDefault(new Locale(((String)setting.getProperty(Key.LANGUAGE))));
				
		if((Boolean)setting.getProperty(Key.IS_FIRST))
		{
			//設定ファイルが存在しない
			setLookAndFeel();

			int ans = 
				JOptionPane.showConfirmDialog(null, 
						Messages.getString("Rflow.first_update_option_check.1") + //$NON-NLS-1$
						NEW_LINE_CODE +
						Messages.getString("Rflow.first_update_option_check.2") + //$NON-NLS-1$
						NEW_LINE_CODE +
						NEW_LINE_CODE +
						Messages.getString("Rflow.first_update_option_check.3") //$NON-NLS-1$
						,
						
						Messages.getString("Message.select"), //$NON-NLS-1$
						JOptionPane.YES_NO_OPTION
						);
			
			setting.setProperty(Key.IS_AUTO_UPDATE, (ans == JOptionPane.YES_OPTION));
			setting.setProperty(Key.IS_FIRST, false);
			
			System.out.println(setting.getProperty(Key.IS_AUTO_UPDATE));
			
			try {
				PropertiesUtility.store(setting, FileResource.SETTING_FILE);
			} catch (IOException e) {
			}
		}
	}
	
	
	
}
