package com.ef_prime.rflow.flow;

import java.awt.geom.Point2D;
import java.util.Map;

import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.UndoableEdit;

import com.ef_prime.rflow.flow.model.FlowEdge;
import com.ef_prime.rflow.flow.model.FlowNode;
import com.ef_prime.rflow.node.NodeModel;
import com.ef_prime.rflow.resource.Messages;

import edu.uci.ics.jung.graph.Vertex;

class UndoableEditFactory {
	private FlowViewer viewer;

	UndoableEditFactory(FlowViewer viewer) {
		this.viewer = viewer;
	}
	
	UndoableEdit createAddNodeEdit(final FlowNode node)
	{
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				viewer.getGraph().removeVertex(node);
				viewer.getPickedState().pick(node, false);
			}
			
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				viewer.getGraph().addVertex(node);
				viewer.getGraphLayout().update();
				viewer.getPickedState().clearPickedVertices();
				viewer.getPickedState().pick(node, true);
				viewer.repaint();
			}
			
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.new_node"); //$NON-NLS-1$
			}
		};
	}

	UndoableEdit createDeleteEdit(
			final FlowNode[] deletedNodes,
			final FlowEdge[] deletedEdges)
	{
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				if(deletedNodes != null)
					for(FlowNode node : deletedNodes)
						viewer.getGraph().addVertex(node);

				if(deletedEdges != null)
					for(FlowEdge edge : deletedEdges)
						viewer.getGraph().addEdge(edge);
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				if(deletedNodes != null)
					for(FlowNode node : deletedNodes)
						viewer.getGraph().removeVertex(node);
				if(deletedNodes == null || deletedNodes.length == 0)
					if(deletedEdges != null)
						for(FlowEdge edge : deletedEdges)
							viewer.getGraph().removeEdge(edge);
			}
			@Override
			public String getPresentationName() {
				if(deletedNodes == null || deletedNodes.length > 0)
					return Messages.getString("Action.remove_node"); //$NON-NLS-1$
				else
					return Messages.getString("Action.remove_edge"); //$NON-NLS-1$
			}
		};
	}
	
	
	UndoableEdit createAddEdgeEdit(final FlowEdge[] edges)
	{
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				for(FlowEdge edge : edges)
					viewer.getGraph().removeEdge(edge);
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				for(FlowEdge edge : edges)
					viewer.getGraph().addEdge(edge);
			}
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.new_edge"); //$NON-NLS-1$
			}
		};
	}
	
	UndoableEdit createAddSublayoutEdit(final FlowNode[] nodes, final FlowEdge[] edges)
	{
		return new AbstractUndoableEdit(){
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				for(FlowNode node : nodes)
					viewer.getGraph().removeVertex(node);
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				for(FlowNode node : nodes)
					viewer.getGraph().addVertex(node);
				for(FlowEdge edge : edges)
					viewer.getGraph().addEdge(edge);
			}
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.add_subflow"); //$NON-NLS-1$
			}
		};
	}
	
	
	UndoableEdit createMoveNodeEdit(final Map<Vertex, Point2D> map){
		return new AbstractUndoableEdit(){
			
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				setLocation();
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				setLocation();
			}
			
			private void setLocation(){
				for(Map.Entry<Vertex, Point2D> entry : map.entrySet()){
					Vertex node = entry.getKey();
					Point2D p = entry.getValue();
					map.put(node, (Point2D)viewer.getGraphLayout().getLocation(node).clone());
					viewer.getGraphLayout().forceMove(node, p.getX(), p.getY());
				}
				viewer.repaint();
			}
			
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.move_node"); //$NON-NLS-1$
			}
		};
	}
	
	
	UndoableEdit createNodeModelChangeEdit(final FlowNode node, final NodeModel oldModel){
		return new AbstractUndoableEdit(){
			
			private NodeModel model = oldModel;
			
			@Override
			public void undo() throws CannotUndoException {
				super.undo();
				switchModel();
			}
			@Override
			public void redo() throws CannotRedoException {
				super.redo();
				switchModel();
			}
			
			private void switchModel(){
				NodeModel temp = node.getModel();
				node.setModel(model);
				model = temp;
				viewer.getPickedState().clearPickedVertices();
				viewer.getPickedState().pick(node, true);
			}
			
			@Override
			public String getPresentationName() {
				return Messages.getString("Action.edit_node"); //$NON-NLS-1$
			}
		}; 
	}
}